/*
 * Copyright (c) 2008, Willow Garage, Inc.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in the
 *       documentation and/or other materials provided with the distribution.
 *     * Neither the name of the Willow Garage, Inc. nor the names of its
 *       contributors may be used to endorse or promote products derived from
 *       this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 *
 */

/**
 * \file 
 * 
 * Interface for storing blobs in db
 *
 * \author Bhaskara Marthi
 */

#ifndef SEMANTICMODEL_BLOB_STORE_H
#define SEMANTICMODEL_BLOB_STORE_H

#include <mongo_ros/message_collection.h>
#include <semanticmodel/blob.hh>
#include <semanticmodel/BlobMessage.h>
#include <semanticmodel/SwitchDB.h>
#include <ros/ros.h>
#include <boost/thread.hpp>
#include <boost/circular_buffer.hpp>
#include <image_transport/image_transport.h>
#include <image_transport/camera_subscriber.h>
#include <image_transport/camera_publisher.h>
#include <image_geometry/pinhole_camera_model.h>
#include <tf/transform_listener.h>
#include <geometry_msgs/Pose2D.h>

namespace semanticmodel
{


class BlobStore
{
public:

  BlobStore (const std::string& canonical_frame,
             const std::string& db_name,
             const std::string& collection_namespace);


  /// \brief Update the database with the current visible set of blobs.
  /// This will add any new blobs (based on id) and leave existing ones
  /// alone.
  void update (const std::vector<Blob*>& blobs,
               const sensor_msgs::Image& img,
               const sensor_msgs::CameraInfo& info);

  /// \brief Switch to using a new db.
  bool switchDb (SwitchDB::Request& req, SwitchDB::Response& resp);

private:

  // Highlight 3d blob in 2d image
  sensor_msgs::Image::Ptr
  highlightBlobInImage (const Blob& blob,
                        const sensor_msgs::Image& img,
                        const sensor_msgs::CameraInfo& info);

  
  // Return the 3d pose of the base frame in the canonical frame
  // at the time that the image was taken
  geometry_msgs::Pose2D getViewpoint (const sensor_msgs::Image& img);

  // How fast is head moving wrt base?
  double headAngularSpeed (const ros::Time& t);

  // Where is head wrt base?
  btQuaternion headAngleAt (const ros::Time& t);

  const string canonical_frame_;
  const string base_frame_;
  const string db_name_;
  const double max_head_angular_speed_;

  boost::mutex mutex_;
  ros::NodeHandle nh_;
  image_transport::ImageTransport it_;

  tf::TransformListener tf_;

  // Camera model for conversion
  image_geometry::PinholeCameraModel cam_model_;

  typedef mongo_ros::MessageCollection<BlobMessage> BlobCollection;
  typedef mongo_ros::MessageCollection<sensor_msgs::Image> ImageCollection;
  
  // DB collection where blobs are stored
  boost::shared_ptr<BlobCollection> blobs_;

  // Collection where images are stored
  boost::shared_ptr<ImageCollection> images_;

  // Publisher for last image
  image_transport::Publisher pub_;

  // Service for switching db
  ros::ServiceServer switch_srv_;
};


} // namespace

#endif // include guard
