/*
 * detailedplane.hh
 * Mac Mason <mmason@willowgarage.com>
 *
 * A plane, as a C++ object; plays nice with semanticmodel::Plane and Planes
 * types.
 */

#pragma once
#include <boost/utility.hpp>
#include "std_msgs/ColorRGBA.h"
#include "tf/transform_listener.h"
#include "pcl/point_types.h"
#include "pcl_ros/point_cloud.h"
#include "sensor_msgs/PointCloud2.h"
#include "semanticmodel/Plane.h"
#include <cstddef>

namespace semanticmodel
{

class DetailedPlane : public boost::noncopyable
{
  private:
    // Save some _serious_ typing.
    typedef pcl::PointXYZRGB Point;
    typedef pcl::PointCloud<Point> PointCloud;

    const static double PLANE_ERROR_TOL = 0.2;

  public:
    // Build one, using a message. The listener parameter is used to transform
    // everything to the /map frame, where we operate.
    DetailedPlane(const semanticmodel::Plane& plane, 
                  tf::TransformListener* listener);

    // No destructor; thanks, shared_ptr!

    // Do these planes overlap, according to our definition?
    bool overlaps(const DetailedPlane& rhs) const;

    // Bring the contents of rhs into this plane; may change very nearly
    // everything. We use the larger of this and rhs to be the "canonical"
    // plane here; that might lead to some goofiness.
    void mergeFrom(const DetailedPlane& rhs);

    // Get a Plane message back out.
    semanticmodel::Plane toPlaneMsg();

    // Like toPlaneMsg, but only fill in the equation and hull.
    void toHullAndEquationPlaneMsg(semanticmodel::Plane& out);

    // The plane equation.
    double a, b, c, d;

    // The centroid of this plane.
    geometry_msgs::Point center;

    // Our convex hull, as a pcl cloud.
    PointCloud::Ptr hull;

    // And all the points that belong to me.
    PointCloud::Ptr cloud;

    // What (false) color am I, for display purposes?
    std_msgs::ColorRGBA color;

    // Global frame
    std::string frame_id;
};

}
