/*
 * Software License Agreement (BSD License)
 *
 *  Copyright (c) 2011 Willow Garage, Inc.
 *    Radu Bogdan Rusu <rusu@willowgarage.com>
 *    Suat Gedikli <gedikli@willowgarage.com>
 *    Patrick Mihelich <mihelich@willowgarage.com>
 *
 *  All rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *   * Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *   * Redistributions in binary form must reproduce the above
 *     copyright notice, this list of conditions and the following
 *     disclaimer in the documentation and/or other materials provided
 *     with the distribution.
 *   * Neither the name of Willow Garage, Inc. nor the names of its
 *     contributors may be used to endorse or promote products derived
 *     from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 *  "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 *  LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 *  FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 *  COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 *  INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 *  BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 *  LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
 *  CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 *  LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
 *  ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 *  POSSIBILITY OF SUCH DAMAGE.
 *
 */
#ifndef OPENNI_RECORD_PLAYER_NODELET_OPENNI_H_
#define OPENNI_RECORD_PLAYER_NODELET_OPENNI_H_

#include <nodelet/nodelet.h>
#include <boost/shared_ptr.hpp>
#include <dynamic_reconfigure/server.h>
#include "semanticmodel/OpenNIConfig.h"
#include <image_transport/image_transport.h>
#include <string>
#include <message_filters/synchronizer.h>
#include <message_filters/subscriber.h>
#include <message_filters/sync_policies/approximate_time.h>
#include <Eigen/Core>
#include <pcl/PointIndices.h>

namespace openni_camera
{
  ////////////////////////////////////////////////////////////////////////////////////////////
  class OpenNIRecordPlayerNodelet : public nodelet::Nodelet
  {
    public:
      virtual ~OpenNIRecordPlayerNodelet ();
    private:
      typedef semanticmodel::OpenNIConfig Config;
      typedef dynamic_reconfigure::Server<Config> ReconfigureServer;
      typedef message_filters::sync_policies::ApproximateTime<sensor_msgs::Image, sensor_msgs::Image> SyncPolicy;
      typedef message_filters::Synchronizer<SyncPolicy> Synchronizer;

      /** \brief Nodelet initialization routine. */
      virtual void onInit ();

      void configCallback (Config &config, uint32_t level);

      // callback methods
      void imageCallback(const sensor_msgs::ImageConstPtr& img_msg);
      void depthCallback(const sensor_msgs::ImageConstPtr& depth_msg);
      void subscriberChangedEvent ();

      // helper methods
      inline bool isImageStreamRequired() const;
      inline bool isDepthStreamRequired() const;
			
			// subscribers
			image_transport::Subscriber img_sub_;
			image_transport::Subscriber depth_sub_;

      // published topics
      image_transport::Publisher pub_depth_image_;
      ros::Publisher pub_disparity_, pub_point_cloud_, pub_point_cloud_rgb_;
      ros::Publisher pub_rgb_cam_info_, pub_depth_cam_info_;
      /** \brief ROS subscribers. */
      ros::Subscriber sub_mask_indices_;
      // Approximate synchronization for XYZRGB point clouds.
      boost::shared_ptr<Synchronizer> depth_rgb_sync_;

			sensor_msgs::CameraInfoPtr fillCameraInfo(const std_msgs::Header &header, unsigned int width, unsigned int height);
      // publish methods
      void publishRgbImage (const sensor_msgs::ImageConstPtr& img_msg) const;
      void publishGrayImage (const sensor_msgs::ImageConstPtr& img_msg) const;
      void publishDepthImage (const sensor_msgs::ImageConstPtr& depth_msg) const;
      void publishDisparity (const sensor_msgs::ImageConstPtr& depth_msg) const;
      void publishXYZPointCloud (const sensor_msgs::ImageConstPtr& depth_msg) const;
      void publishXYZRGBPointCloud (const sensor_msgs::ImageConstPtr& depth_msg, const sensor_msgs::ImageConstPtr& rgb_msg) const;

      /** \brief Store provided indices for later use in masking the point cloud */
      void maskIndicesCb(const pcl::PointIndicesConstPtr& indices);

      /** \brief reconfigure server*/
      boost::shared_ptr<ReconfigureServer> reconfigure_server_;
      Config config_;
      boost::recursive_mutex reconfigure_mutex_;
      //boost::mutex depth_mutex_;
      //boost::mutex image_mutex_;

      /** \brief in case didnot recorded camera info */
      bool generate_camera_info_;
      /** \brief whether to listen for mask_indices */
      bool use_indices_;
      /** \brief vector to hold the 'mask' (indices to output) **/
      std::vector<int32_t> mask_indices_;
    public:
      EIGEN_MAKE_ALIGNED_OPERATOR_NEW;
  };

  bool OpenNIRecordPlayerNodelet::isImageStreamRequired() const
  {
    return (pub_point_cloud_rgb_.getNumSubscribers() > 0 );
  }

  bool OpenNIRecordPlayerNodelet::isDepthStreamRequired() const
  {
    return (pub_depth_image_.getNumSubscribers()     > 0 ||
            pub_disparity_.getNumSubscribers()       > 0 ||
            pub_point_cloud_.getNumSubscribers()     > 0 ||
            pub_point_cloud_rgb_.getNumSubscribers() > 0 );
  }

}
#endif  //#ifndef OPENNI_NODELET_RECORD_PLAYER_OPENNI_H_
