/*
 * segmenter.hh
 * Mac Mason <mmason@willowgarage.com>
 *
 * Segment out planes & objects & whatnot. 
 *
 * This object needs two things as input: a plane (PointCloud) and an Image.
 * It outputs several things:
 *
 * - A pointcloud of the current plane. 
 * - Planes, as Plane messages.
 * - Objects, as PointClouds.
 *
 * This object maintains all of its state internally; the NodeHandle, the
 * publishers and subscribers, and everything else. So, build one, configure
 * it, start it up, and spin.
 */

#pragma once

#include <string>
#include <boost/utility.hpp>  // For noncopyable
#include <boost/scoped_ptr.hpp> 
#include <boost/thread.hpp>

#include "ros/ros.h"

#include "pcl/point_types.h"
#include "pcl_ros/point_cloud.h"
#include "pcl/ModelCoefficients.h"
#include "tf/transform_listener.h"
#include "tf/transform_broadcaster.h"
#include "image_transport/image_transport.h"
#include "sensor_msgs/CameraInfo.h"
#include "message_filters/synchronizer.h"
#include "message_filters/sync_policies/approximate_time.h"

#include "dynamic_reconfigure/server.h"
#include "semanticmodel/FilterConfig.h"
#include "semanticmodel/Plane.h"
#include <cstddef>

using semanticmodel::FilterConfig;
using std::vector;

namespace semanticmodel
{
  class Segmenter : public boost::noncopyable
  {
    private:
      // Save some _serious_ typing.
      typedef pcl::PointXYZRGB Point;
      typedef pcl::PointCloud<Point> PointCloud;

      // Time here is tricky. However, openni_camera does approximate
      // synchronization for us, and then sets the cloud's timestamp to the
      // depth image's stamp. Sadly, that isn't (necessarily) the same as the
      // RGB image's stamp. Instead, we approximately synchronize all three,
      // knowing that (we hope) the first two are already exactly
      // synchronized, and the third has already passed through an
      // ApproximateTime policy.
      typedef message_filters::sync_policies::
        ApproximateTime<sensor_msgs::Image, 
                        sensor_msgs::CameraInfo, 
                        PointCloud> Sync;

    public:
      // Any of the output topics can be the empty string, in which case we
      // don't publish them. The input topics, however, must both be provided.
      //
      // camera_input_topic is the (partial) namespace of the camera. For
      // example, the Kinect publishes /camera/rgb/image_color and
      // /camera/rgb/camera_info; camera_input_topic should be /camera/rgb.
      // (camera_info and image_color are hardcoded.)
      //
      // The passed-in topic names are appended to name_space (this permits
      // private naming).
      Segmenter(const std::string& name_space,
                const std::string& cloud_input_topic,
                const std::string& camera_input_topic,
                const std::string& planecloud_output_topic,
                const std::string& plane_output_topic,
                const std::string& object_output_topic,
                const std::string& plane_service_topic);

      // No need for a destructor; our members are smart. No need for other
      // stuff; noncopyable keeps us well-behaved.
      
      /*
       * Our processing pipeline. This has several steps:
       *
       * 0. Transform the input into the config.canonical_frame tf frame;
       * this is (probably) /map, but could generally be any trustworthy frame
       * where +z is "up". This allows us to think about "horizontal" in very
       * simple terms.
       *
       * 1. A passthrough filter. This filters out NaNs, but (more
       * importantly) truncates out the floor and ceiling.
       *
       * 2. A VoxelGrid filter. Dealing with the cloud at full resolution
       * (particularly for plane extraction) is painfully slow; we use the
       * downsampled cloud to speed stuff up.
       *
       * 3. A plane fitter (using RANSAC). The key here is that this looks for
       * _horizontal_ planes, not just any planes; this means it's going to be
       * running a loop, so this is likely to be the slow part. Furthermore,
       * at this point, the pipeline gets wider, because we might find more
       * than one valid, horizontal, plane per cloud. The output from this
       * function is zero or more planes, as PointClouds, already projected
       * onto their plane equations, and a vector of ModelCoefficients
       * objects, representing those equations.
       *
       * 4. Erosion. Plane fitting finds infinite planes, which means that
       * tables and things also end up including little bits of faraway walls
       * and the like; this uses some neighbor-search operations to get rid of
       * them. It's named "erosion" after the morphological operator that does
       * something similar. (In fact, actually using erosion might be a better
       * way to do this step.) TODO TODO: Is it really?
       *
       * 5. In-plane clustering. An infinite plane (even after erosion) can
       * include parts of several different planes as the same height; we
       * cluster them out, so that we get only one plane at a time. 
       *
       * 6. Get the convex hull of the extracted plane. 
       *
       * 7. The convex hulls we've just built aren't quite big enough; we want
       * to search for objects atop those hulls merged with existing ones
       * (which is what plane_tracker does). So, send them to the tracker, and
       * get back the hulls to use for the next step.
       *
       * 8. Collect the points above the convex hull; these are potential
       * objects.
       *
       * 9. Cluster those points to generate actual objects.
       *
       * These functions are public, for the moment, because they might be
       * generally useful, or something. They might not really deserve
       * publicity.
       */
      void passthrough(PointCloud::Ptr& in, PointCloud::Ptr& out);
      void voxelize(PointCloud::Ptr& in, PointCloud::Ptr& out);

      // fit_planes is going to edit both in and out. Be aware. good_out and
      // good_models are the horizontal planes (and their equations), whereas
      // bad_models stores the equations of the non-horizontal planes.
      void fit_planes(PointCloud::Ptr& in, 
                     vector<PointCloud::Ptr>& good_out,
                     vector<pcl::ModelCoefficients::Ptr>& good_models,
                     vector<pcl::ModelCoefficients::Ptr>& bad_models);

      // planes is both an input and output argument.
      void erode(vector<PointCloud::Ptr>& planes);

      // planes is both an input and an output argument, as is coeffs. This
      // function can completely alter the contents of both vectors; watch
      // out. 
      void plane_clusters(vector<PointCloud::Ptr>& planes,
                          vector<pcl::ModelCoefficients::Ptr>& coeffs);

      // We need the contents of both planes and hulls, so planes isn't
      // modified here.
      void convex_hulls(const vector<PointCloud::Ptr>& planes,
                        vector<PointCloud::Ptr>& hulls);

      // Send in some information, get back out new planes.
      void publish_planes(
          const vector<PointCloud::Ptr>& planes,
          const vector<PointCloud::Ptr>& hulls,
          const vector<pcl::ModelCoefficients::Ptr>& coeffs,
          vector<semanticmodel::Plane>& outputs);

      // Note that we start working with the un-voxelized cloud again.
      void get_aboves(const vector<PointCloud::Ptr>& hulls,
                      const vector<pcl::ModelCoefficients::Ptr>& coeffs,
                      const PointCloud::Ptr& densecloud,
                      PointCloud::Ptr& out);

      // Throw out those "above" points that fit one of the bad_coeffs planes
      // too well. This helps us get rid of extraneous "wall" objects.
      void lose_bads(const PointCloud::ConstPtr& aboves,
                     PointCloud::Ptr good_aboves,
                     const vector<pcl::ModelCoefficients::Ptr>& bad_coeffs);

      // Finally, the callback that actually drives everything.
      void synchronized_pipeline_callback(
          const sensor_msgs::ImageConstPtr& img,
          const sensor_msgs::CameraInfoConstPtr& info,
          const PointCloud::ConstPtr& cloud);

    private:
      // How big are our queues?
      static const int QUEUE_SIZE = 100;

      // We are a node.
      ros::NodeHandle handle;

      // For dynamic_reconfigure support.
      FilterConfig config;
      dynamic_reconfigure::Server<FilterConfig> dynparam_srv;
      dynamic_reconfigure::Server<FilterConfig>::CallbackType dynparam_func;
      boost::mutex config_mutex;

      // Finally, the dynamic_reconfigure callback.
      void dynamic_reconfigure_callback(const FilterConfig& config, 
                                        uint32_t level);
      
      // tf support; pointers are needed here to enforce that we have a
      // nodehandle before these are built.
      boost::scoped_ptr<tf::TransformListener> listener;
      boost::scoped_ptr<tf::TransformBroadcaster> broadcaster;

      // Squeeze all of our messages together. See the long comment below,
      // too.
      boost::scoped_ptr< message_filters::Synchronizer<Sync> > sync;

      /*
       * We need image_transport to play nicely with ApproximateTime, which it
       * doesn't (automatically), because the ApproximateTime stuff requires
       * message_filter subscriptions, not image_transport subscriptions. We
       * can get around this (cf openni_record_player) using .add<i> on our
       * synchronizer; this means we need barely-op callbacks for the three
       * channels we're watching, plus subscribers for them.
       */
      ros::Subscriber cloud_sub;
      image_transport::ImageTransport transport; // For the next one.
      image_transport::CameraSubscriber cam_sub; // Gets both.

      // The barely-op callbacks described above; these just push their
      // messages into the sync queues.
      void cloud_cb(const PointCloud::ConstPtr& cloud);
      void camera_cb(const sensor_msgs::ImageConstPtr& img, 
                     const sensor_msgs::CameraInfoConstPtr& info);

      // The various things we push out.
      ros::Publisher planecloud_pub;
      ros::Publisher plane_pub;
      ros::Publisher object_pub;

      // For our plane-merging service
      ros::ServiceClient planetracker_srv;

      // If we're on a robot, we can short-circuit early (as soon as planes
      // are published for the benefit of the head pointer). This is collected
      // from a param at construction-time.
      bool short_circuit;
      
      // Some debugging data.
      int processed;
      int skipped;
  };
}
