/*
 * Software License Agreement (BSD License)
 *
 *  Copyright (c) 2011 Willow Garage, Inc.
 *    Suat Gedikli <gedikli@willowgarage.com>
 *    Patrick Michelich <michelich@willowgarage.com>
 *    Radu Bogdan Rusu <rusu@willowgarage.com>
 *
 *  All rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *   * Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *   * Redistributions in binary form must reproduce the above
 *     copyright notice, this list of conditions and the following
 *     disclaimer in the documentation and/or other materials provided
 *     with the distribution.
 *   * Neither the name of Willow Garage, Inc. nor the names of its
 *     contributors may be used to endorse or promote products derived
 *     from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 *  "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 *  LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 *  FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 *  COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 *  INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 *  BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 *  LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
 *  CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 *  LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
 *  ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 *  POSSIBILITY OF SUCH DAMAGE.
 *
 */
#include <pluginlib/class_list_macros.h>
#include "semanticmodel/openni_record_player_nodelet.h"
#include "openni_camera/openni_image.h"
#include "openni_camera/openni_depth_image.h"
#include <sensor_msgs/image_encodings.h>
#include <pcl_ros/point_cloud.h>
#include <pcl/point_types.h>
#include <stereo_msgs/DisparityImage.h>
#include <cv_bridge/cv_bridge.h>
#include "opencv/cv.h"
#include "opencv/highgui.h"

#if ROS_VERSION_MINIMUM(1, 3, 0)
#include <sensor_msgs/distortion_models.h>
#endif

using namespace std;
namespace openni_camera
{
inline bool operator == (const XnMapOutputMode& mode1, 
                         const XnMapOutputMode& mode2)
{
  return (mode1.nXRes == mode2.nXRes && mode1.nYRes == mode2.nYRes && mode1.nFPS == mode2.nFPS);
}
inline bool operator != (const XnMapOutputMode& mode1, 
                         const XnMapOutputMode& mode2)
{
  return !(mode1 == mode2);
}


PLUGINLIB_DECLARE_CLASS (openni_camera, 
                         OpenNIRecordPlayerNodelet, 
                         openni_camera::OpenNIRecordPlayerNodelet, 
                         nodelet::Nodelet);

typedef union
{
  struct /*anonymous*/
  {
    unsigned char Blue;
    unsigned char Green;
    unsigned char Red;
    unsigned char Alpha;
  };
  float float_value;
  long long_value;
} RGBValue;

#define KINECT_FOCAL		525
#define KINECT_BASELINE 0.075

OpenNIRecordPlayerNodelet::~OpenNIRecordPlayerNodelet ()
{
}

void OpenNIRecordPlayerNodelet::onInit ()
{
  std::string camera_name(getNodeHandle ().resolveName ("camera"));
  ros::NodeHandle comm_nh(camera_name); // for topics, services
  ros::NodeHandle param_nh = getPrivateNodeHandle (); // for parameters

  param_nh.param ("use_indices", use_indices_, false);
  // setup image subscription call back
  std::string image_topic("/" + camera_name + "/rgb/image_color");
  std::string depth_topic("/" + camera_name + "/depth/image_raw");


  image_transport::ImageTransport imageTransport (comm_nh);

  img_sub_ = imageTransport.subscribe(image_topic, 2, &OpenNIRecordPlayerNodelet::imageCallback, this);
  depth_sub_ = imageTransport.subscribe(depth_topic, 2, &OpenNIRecordPlayerNodelet::depthCallback, this);

  image_transport::SubscriberStatusCallback subscriberChanged = boost::bind(&OpenNIRecordPlayerNodelet::subscriberChangedEvent, this);
  pub_depth_image_ = imageTransport.advertise ("depth/image"    , 5, subscriberChanged, subscriberChanged );
  ros::SubscriberStatusCallback subscriberChanged2 = boost::bind(&OpenNIRecordPlayerNodelet::subscriberChangedEvent, this);
  pub_disparity_ = comm_nh.advertise<stereo_msgs::DisparityImage > ("depth/disparity", 5, subscriberChanged2, subscriberChanged2);
  pub_point_cloud_ = comm_nh.advertise<pcl::PointCloud<pcl::PointXYZ> > ("depth/points", 5, subscriberChanged2, subscriberChanged2);
  pub_point_cloud_rgb_ = comm_nh.advertise<pcl::PointCloud<pcl::PointXYZRGB> > ("rgb/points", 5, subscriberChanged2, subscriberChanged2);

	generate_camera_info_ = 0;  // Mac says: I don't want this, I think.
	if(generate_camera_info_) {
		pub_rgb_cam_info_ = comm_nh.advertise<sensor_msgs::CameraInfo > ("rgb/camera_info", 5, subscriberChanged2, subscriberChanged2);
		pub_depth_cam_info_ = comm_nh.advertise<sensor_msgs::CameraInfo > ("depth/camera_info", 5, subscriberChanged2, subscriberChanged2);
	}

  if (use_indices_)
    sub_mask_indices_ = comm_nh.subscribe("depth/indices", 5, &OpenNIRecordPlayerNodelet::maskIndicesCb, this);

  SyncPolicy sync_policy (4); // queue size
  depth_rgb_sync_.reset (new Synchronizer (sync_policy));
  depth_rgb_sync_->registerCallback (boost::bind (&OpenNIRecordPlayerNodelet::publishXYZRGBPointCloud, this, _1, _2));

	ROS_WARN("Using hard-coded focal length(%d)/baseline value(%f). may cause a problem", KINECT_FOCAL, KINECT_BASELINE);
}

sensor_msgs::CameraInfoPtr OpenNIRecordPlayerNodelet::fillCameraInfo(const std_msgs::Header &header, unsigned int width, unsigned int height)
{
  sensor_msgs::CameraInfoPtr info_msg = boost::make_shared<sensor_msgs::CameraInfo > ();
  info_msg->header  		  	= header;
  info_msg->width           = width;
  info_msg->height          = height;
#if ROS_VERSION_MINIMUM(1, 3, 0)
  info_msg->D = std::vector<double>(5, 0.0);
  info_msg->distortion_model = sensor_msgs::distortion_models::PLUMB_BOB;
#else
  info_msg->D.assign (0.0);
#endif
  info_msg->K.assign (0.0);
  info_msg->R.assign (0.0);
  info_msg->P.assign (0.0);

  // Simple camera matrix: square pixels, principal point at center
  info_msg->K[0] = info_msg->K[4] = KINECT_FOCAL;
  info_msg->K[2] = (info_msg->width / 2) - 0.5;
  info_msg->K[5] = (info_msg->width * 3./8.) - 0.5; //aspect ratio for the camera center on kinect and presumably other devices is 4/3
  info_msg->K[8] = 1.0;
  // no rotation: identity
  info_msg->R[0] = info_msg->R[4] = info_msg->R[8] = 1.0;
  // no rotation, no translation => P=K(I|0)=(K|0)
  info_msg->P[0] = info_msg->P[5] = info_msg->K[0];
  info_msg->P[2] = info_msg->K[2];
  info_msg->P[6] = info_msg->K[5];
  info_msg->P[10] = 1.0;

  return info_msg;
}

void OpenNIRecordPlayerNodelet::imageCallback (const sensor_msgs::ImageConstPtr& img_msg)
{
#if 0
	assert(img_msg->encoding == sensor_msgs::image_encodings::BAYER_GRBG8); // make sure the messages are raw format
	if (pub_rgb_image_.getNumSubscribers () > 0 
		|| pub_point_cloud_rgb_.getNumSubscribers () > 0 )
    publishRgbImage(img_msg);

  if (pub_gray_image_.getNumSubscribers () > 0)
    publishGrayImage(img_msg);
#else
	assert(img_msg->encoding == sensor_msgs::image_encodings::BGR8); // make sure the messages are raw format
  if (pub_point_cloud_rgb_.getNumSubscribers () > 0)
    depth_rgb_sync_->add < 1 > (img_msg);
	
	if(generate_camera_info_) pub_rgb_cam_info_.publish (fillCameraInfo (img_msg->header, (unsigned int)img_msg->width, (unsigned int)img_msg->height));
#endif
}

void OpenNIRecordPlayerNodelet::depthCallback (const sensor_msgs::ImageConstPtr& depth_msg)
{
  if (pub_depth_image_.getNumSubscribers () > 0 || pub_point_cloud_rgb_.getNumSubscribers () > 0)
		publishDepthImage(depth_msg);
  if (pub_disparity_.getNumSubscribers () > 0)
		publishDisparity(depth_msg);
	if(pub_point_cloud_.getNumSubscribers() > 0)
		publishXYZPointCloud(depth_msg);

	if(generate_camera_info_) pub_depth_cam_info_.publish (fillCameraInfo (depth_msg->header, (unsigned int)depth_msg->width, (unsigned int)depth_msg->height));
}

void OpenNIRecordPlayerNodelet::subscriberChangedEvent ()
{
}

// don't need... just run ROS_NAMESPACE=camera/rgb rosrun image_proc image_proc
#if 0
void OpenNIRecordPlayerNodelet::publishRgbImage (const sensor_msgs::ImageConstPtr& img_msg) const
{
#if 0 
	// CV_bridge doesnot support image conversion from bayer...
	// cannot use this..

	cv_bridge::CvImagePtr cv_ptr = cv_bridge::toCvCopy(img_msg, std::string("bgr8"));
  sensor_msgs::ImagePtr rgb_msg = cv_ptr->toImageMsg();
  rgb_msg->header.stamp = img_msg->header.stamp;
	assert(rgb_msg->encoding != sensor_msgs::image_encodings::BAYER_GRBG8); // make sure the messages are raw format
#else
  sensor_msgs::ImagePtr rgb_msg = boost::make_shared<sensor_msgs::Image > ();
  rgb_msg->header.stamp = img_msg->header.stamp;
  rgb_msg->header.frame_id = img_msg->header.frame_id;
  rgb_msg->encoding = sensor_msgs::image_encodings::RGB8;
  rgb_msg->height = img_msg->height;
  rgb_msg->width = img_msg->width;
  rgb_msg->step = img_msg->width * 3;
  rgb_msg->data.resize (rgb_msg->height * rgb_msg->step);
	// convert image to rgb from raw format.
  // image.fillRGB (rgb_msg->width, rgb_msg->height, &rgb_msg->data[0], rgb_msg->step);
#endif
  if (pub_rgb_image_.getNumSubscribers () > 0)
    pub_rgb_image_.publish (rgb_msg);
  if (pub_point_cloud_rgb_.getNumSubscribers () > 0)
    depth_rgb_sync_->add < 1 > (rgb_msg);
}

void OpenNIRecordPlayerNodelet::publishGrayImage (const sensor_msgs::ImageConstPtr& img_msg) const
{
  sensor_msgs::ImagePtr gray_msg = boost::make_shared<sensor_msgs::Image > ();
  gray_msg->header.stamp = img_msg->header.stamp;
  gray_msg->header.frame_id = img_msg->header.frame_id;
  gray_msg->encoding = sensor_msgs::image_encodings::MONO8;
  gray_msg->height = img_msg->height;
  gray_msg->width = img_msg->width;
  gray_msg->step = img_msg->width;
  gray_msg->data.resize (gray_msg->height * gray_msg->step);
	// convert image to rgb from raw format.
  // image.fillGrayscale (gray_msg->width, gray_msg->height, &gray_msg->data[0], gray_msg->step);
  pub_gray_image_.publish (gray_msg);
}
#endif

void OpenNIRecordPlayerNodelet::publishDepthImage (const sensor_msgs::ImageConstPtr& depth_raw_msg) const
{
  sensor_msgs::ImagePtr depth_msg = boost::make_shared<sensor_msgs::Image > ();
  depth_msg->header.stamp         = depth_raw_msg->header.stamp;
  depth_msg->header.frame_id      = depth_raw_msg->header.frame_id;
  depth_msg->encoding             = sensor_msgs::image_encodings::TYPE_32FC1;
  depth_msg->height               = depth_raw_msg->height;
  depth_msg->width                = depth_raw_msg->width;
  depth_msg->step                 = depth_msg->width * sizeof (float);
  depth_msg->data.resize (depth_msg->height * depth_msg->step);

	float *depth_buffer = reinterpret_cast<float*>((void*)&depth_msg->data[0]);
	short *raw_buffer = reinterpret_cast<short*>((void*)&depth_raw_msg->data[0]);
  
  for (unsigned yIdx = 0; yIdx < depth_msg->height; ++yIdx)
    for (unsigned xIdx = 0; xIdx < depth_msg->width; ++xIdx, ++depth_buffer, ++raw_buffer)
			*depth_buffer = (float)*raw_buffer * 0.001f; // mm to m
	
	pub_depth_image_.publish (depth_msg);
  if (pub_point_cloud_rgb_.getNumSubscribers () > 0)
    depth_rgb_sync_->add < 0 > (depth_msg);
}

void OpenNIRecordPlayerNodelet::publishDisparity (const sensor_msgs::ImageConstPtr& depth_raw_msg) const
{
  stereo_msgs::DisparityImagePtr disp_msg = boost::make_shared<stereo_msgs::DisparityImage > ();
  disp_msg->header.stamp                  = depth_raw_msg->header.stamp;
  disp_msg->header.frame_id               = depth_raw_msg->header.frame_id;
  disp_msg->image.header                  = disp_msg->header;
  disp_msg->image.encoding                = sensor_msgs::image_encodings::TYPE_32FC1;
  disp_msg->image.height                  = depth_raw_msg->height;
  disp_msg->image.width                   = depth_raw_msg->width;
  disp_msg->image.step                    = disp_msg->image.width * sizeof (float);
  disp_msg->image.data.resize (disp_msg->image.height * disp_msg->image.step);

	/* FIX ME : extracted from openi_nodelet.cpp */
  disp_msg->T = KINECT_BASELINE; // depth_raw_msg->T; // baseline
  disp_msg->f = KINECT_FOCAL; // depth_raw_msg->f; // f
  disp_msg->min_disparity = 0.0;
  disp_msg->max_disparity = disp_msg->T * disp_msg->f / 0.3;
  disp_msg->delta_d = 0.125;
	/*===========================================*/

	float *disp_buffer = reinterpret_cast<float*>((void*)&disp_msg->image.data[0]);
	short *raw_buffer = reinterpret_cast<short*>((void*)&depth_raw_msg->data[0]);
  
	float constant = disp_msg->f * disp_msg->T * 1000.0;
  for (unsigned yIdx = 0; yIdx < depth_raw_msg->height; ++yIdx)
    for (unsigned xIdx = 0; xIdx < depth_raw_msg->width; ++xIdx, ++disp_buffer, ++raw_buffer)
			*disp_buffer = constant / (double)*raw_buffer;

  pub_disparity_.publish (disp_msg);
}

void OpenNIRecordPlayerNodelet::publishXYZPointCloud (const sensor_msgs::ImageConstPtr& depth_raw_msg) const
{
  pcl::PointCloud<pcl::PointXYZ>::Ptr cloud_msg (new pcl::PointCloud<pcl::PointXYZ>() );
  cloud_msg->header.stamp = depth_raw_msg->header.stamp;
	cloud_msg->header.frame_id = depth_raw_msg->header.frame_id;

  bool do_masking = (use_indices_ and (mask_indices_.size() != 0));
  float centerX, centerY;

  if (not do_masking)
  {
    cloud_msg->height = depth_raw_msg->height;
    cloud_msg->width = depth_raw_msg->width;
    centerX = (cloud_msg->width >> 1) - 0.5f;
    centerY = (cloud_msg->height >> 1) - 0.5f;
  }
  else
  {
    cloud_msg->width = mask_indices_.size();
    cloud_msg->height = 1;
    centerX = (depth_raw_msg->width >> 1) - 0.5f;
    centerY = (depth_raw_msg->height >> 1) - 0.5f;
  }
  cloud_msg->is_dense = false;

  cloud_msg->points.resize (cloud_msg->height * cloud_msg->width);
  float constant = 0.001 / KINECT_FOCAL; // device_->getDepthFocalLength (depth_width_);

	short *raw_buffer = reinterpret_cast<short*>((void*)&depth_raw_msg->data[0]);
  float bad_point = std::numeric_limits<float>::quiet_NaN ();
  int depth_idx = 0;

  pcl::PointCloud<pcl::PointXYZ>::iterator pt_iter = cloud_msg->begin ();
  if (not do_masking)
  {
    for (int v = 0; v < (int)cloud_msg->height; ++v)
    {
      for (int u = 0; u < (int)cloud_msg->width; ++u, raw_buffer++, ++pt_iter)
      {
        pcl::PointXYZ& pt = *pt_iter;

        // Check for invalid measurements
        if (std::isnan(*(raw_buffer)))
        {
          pt.x = pt.y = pt.z = bad_point;
          continue;
        }

        // Fill in XYZ
        pt.x = (u - centerX) * (*raw_buffer) * constant;
        pt.y = (v - centerY) * (*raw_buffer) * constant;
        pt.z = (*raw_buffer) * 0.001;
      }
    }
  }
  else
  {
    int u, v;
    unsigned int nMask = mask_indices_.size();
    int max_index = (depth_raw_msg->height * depth_raw_msg->width) - 1;
    for (unsigned int i = 0; i < nMask; i++, ++pt_iter)
    {
      pcl::PointXYZ& pt = *pt_iter;
      depth_idx = mask_indices_[i];
      if (depth_idx > max_index)
      {
        NODELET_ERROR("Mask index %d exceeds maximum index of %d", depth_idx, max_index);
        continue;
      }
      v = depth_idx / depth_raw_msg->width;
      u = depth_idx % depth_raw_msg->width;
      // Check for invalid measurements
      if (std::isnan(raw_buffer[depth_idx]))
      {
        // not valid
        pt.x = pt.y = pt.z = bad_point;
        continue;
      }
      // Fill in XYZ
      pt.x = (u - centerX) * raw_buffer[depth_idx] * constant;
      pt.y = (v - centerY) * raw_buffer[depth_idx] * constant;
      pt.z = raw_buffer[depth_idx] * 0.001;
    }
  }
  pub_point_cloud_.publish (cloud_msg);
}

void OpenNIRecordPlayerNodelet::publishXYZRGBPointCloud (const sensor_msgs::ImageConstPtr& depth_msg, const sensor_msgs::ImageConstPtr& rgb_msg) const
{
  pcl::PointCloud<pcl::PointXYZRGB>::Ptr cloud_msg (new pcl::PointCloud<pcl::PointXYZRGB>() );
  cloud_msg->header.stamp     = depth_msg->header.stamp;
  cloud_msg->header.frame_id  = rgb_msg->header.frame_id;
  cloud_msg->is_dense         = false;

  bool do_masking = (use_indices_ and (mask_indices_.size() != 0));

  float centerX, centerY;
  float constant;
  unsigned color_step, color_skip;

  if (not do_masking)
  {
    cloud_msg->height = depth_msg->height;
    cloud_msg->width = depth_msg->width;
    centerX = (cloud_msg->width >> 1) - 0.5f;
    centerY = (cloud_msg->height >> 1) - 0.5f;
    constant = 1.0f / KINECT_FOCAL; // device_->getImageFocalLength(cloud_msg->width);
    color_step = 3 * rgb_msg->width / cloud_msg->width;
    color_skip = 3 * (rgb_msg->height / cloud_msg->height - 1) * rgb_msg->width;

  }
  else
  {
    cloud_msg->width = mask_indices_.size();
    cloud_msg->height = 1;
    centerX = (depth_msg->width >> 1) - 0.5f;
    centerY = (depth_msg->height >> 1) - 0.5f;
    constant = 1.0f / KINECT_FOCAL; //device_->getImageFocalLength(depth_msg->width);
    color_step = 3 * rgb_msg->width / depth_msg->width;
    color_skip = 3 * (rgb_msg->height / depth_msg->height - 1) * rgb_msg->width;
 }

  static bool info_sent = false;
  if (not info_sent)
  {
    //NODELET_INFO("color_step = %d, color_skip = %d", color_step, color_skip);
    info_sent = true;
  }

  // do not publish if rgb image is smaller than color image -> seg fault
  if (rgb_msg->height < depth_msg->height || rgb_msg->width < depth_msg->width)
  {
    // we dont want to flood the terminal with warnings
    static unsigned warned = 0;
    if (warned % 100 == 0)
      NODELET_WARN("rgb image smaller than depth image... skipping point cloud for this frame rgb:%dx%d vs. depth:%3dx%d"
              , rgb_msg->width, rgb_msg->height, depth_msg->width, depth_msg->height );
    ++warned;
    return;
  }
  cloud_msg->points.resize (cloud_msg->height * cloud_msg->width);

  const float* depth_buffer = reinterpret_cast<const float*>(&depth_msg->data[0]);
  const uint8_t* rgb_buffer = &rgb_msg->data[0];

  // depth_msg already has the desired dimensions, but rgb_msg may be higher res.
  int color_idx = 0, depth_idx = 0;
  pcl::PointCloud<pcl::PointXYZRGB>::iterator pt_iter = cloud_msg->begin ();
  if (not do_masking)
  {
    for (int v = 0; v < (int)cloud_msg->height; ++v, color_idx += color_skip)
    {
      for (int u = 0; u < (int)cloud_msg->width; ++u, color_idx += color_step, ++depth_idx, ++pt_iter)
      {
        pcl::PointXYZRGB& pt = *pt_iter;
        float Z = depth_buffer[depth_idx];

        // Check for invalid measurements
        if (std::isnan (Z))
        {
          pt.x = pt.y = pt.z = Z;
        }
        else
        {
          // Fill in XYZ
          pt.x = (u - centerX) * Z * constant;
          pt.y = (v - centerY) * Z * constant;
          pt.z = Z;
        }

        // Fill in color
        RGBValue color;
				// using BGR format...
        color.Red   = rgb_buffer[color_idx + 2];
        color.Green = rgb_buffer[color_idx + 1];
        color.Blue  = rgb_buffer[color_idx];
        color.Alpha = 0;
        pt.rgb = color.float_value;
      }
    }
  }
  else
  {
    int u, v;
    unsigned int nMask = mask_indices_.size();
    int max_index = (depth_msg->height * depth_msg->width) - 1;
    for (unsigned int i = 0; i < nMask; i++, ++pt_iter)
    {
      pcl::PointXYZRGB& pt = *pt_iter;
      depth_idx = mask_indices_[i];
      if (depth_idx > max_index)
      {
        NODELET_ERROR("Mask index %d exceeds maximum index of %d", depth_idx, max_index);
        continue;
      }
      float Z = depth_buffer[depth_idx];

      v = depth_idx / depth_msg->width;
      u = depth_idx % depth_msg->width;
      // Check for invalid measurements

      if (std::isnan (Z))
      {
        pt.x = pt.y = pt.z = Z;
      }
      else
      {
        // Fill in XYZ
        pt.x = (u - centerX) * Z * constant;
        pt.y = (v - centerY) * Z * constant;
        pt.z = Z;
      }

      // Fill in color
      RGBValue color;
      color_idx = (v*depth_msg->width + u)*color_step;
			// using BGR format...
      color.Red   = rgb_buffer[color_idx + 2];
      color.Green = rgb_buffer[color_idx + 1];
      color.Blue  = rgb_buffer[color_idx ];
      color.Alpha = 0;
      pt.rgb = color.float_value;
    }
  }

  pub_point_cloud_rgb_.publish (cloud_msg);
}

void OpenNIRecordPlayerNodelet::configCallback (Config &config, uint32_t level)
{
  config_ = config;
}

void OpenNIRecordPlayerNodelet::maskIndicesCb(const pcl::PointIndicesConstPtr& indices)
{
  mask_indices_.resize(indices->indices.size());
//  NODELET_INFO("New mask with %d indices", int(mask_indices_.size()));
  memcpy(mask_indices_.data(), indices->indices.data(), indices->indices.size() * sizeof(int32_t));
}
}


