/*
 * Software License Agreement (BSD License)
 *
 *  Copyright (c) 2011 Willow Garage, Inc.
 *    Suat Gedikli <gedikli@willowgarage.com>
 *    Patrick Michelich <michelich@willowgarage.com>
 *    Radu Bogdan Rusu <rusu@willowgarage.com>
 *
 *  All rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *   * Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *   * Redistributions in binary form must reproduce the above
 *     copyright notice, this list of conditions and the following
 *     disclaimer in the documentation and/or other materials provided
 *     with the distribution.
 *   * Neither the name of Willow Garage, Inc. nor the names of its
 *     contributors may be used to endorse or promote products derived
 *     from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 *  "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 *  LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 *  FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 *  COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 *  INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 *  BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 *  LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
 *  CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 *  LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
 *  ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 *  POSSIBILITY OF SUCH DAMAGE.
 *
 */
#include <pluginlib/class_list_macros.h>
#include "openni_camera/openni_record_player_nodelet.h"
#include "openni_camera/openni_image.h"
#include "openni_camera/openni_depth_image.h"
#include <sensor_msgs/image_encodings.h>
#include <pcl_ros/point_cloud.h>
#include <pcl/point_types.h>
#include <stereo_msgs/DisparityImage.h>

#if ROS_VERSION_MINIMUM(1, 3, 0)
#include <sensor_msgs/distortion_models.h>
#endif

using namespace std;
namespace openni_camera
{
inline bool operator == (const XnMapOutputMode& mode1, const XnMapOutputMode& mode2)
{
  return (mode1.nXRes == mode2.nXRes && mode1.nYRes == mode2.nYRes && mode1.nFPS == mode2.nFPS);
}
inline bool operator != (const XnMapOutputMode& mode1, const XnMapOutputMode& mode2)
{
  return !(mode1 == mode2);
}


PLUGINLIB_DECLARE_CLASS (openni_camera, OpenNIRecordPlayerNodelet, openni_camera::OpenNIRecordPlayerNodelet, nodelet::Nodelet);

typedef union
{
  struct /*anonymous*/
  {
    unsigned char Blue;
    unsigned char Green;
    unsigned char Red;
    unsigned char Alpha;
  };
  float float_value;
  long long_value;
} RGBValue;

OpenNIRecordPlayerNodelet::~OpenNIRecordPlayerNodelet ()
{
}

void OpenNIRecordPlayerNodelet::onInit ()
{
  ros::NodeHandle comm_nh(getNodeHandle ().resolveName ("camera")); // for topics, services
  ros::NodeHandle param_nh = getPrivateNodeHandle (); // for parameters

  param_nh.param ("rgb_frame_id", rgb_frame_id_, string (""));
  if (rgb_frame_id_.empty ())
  {
    rgb_frame_id_ = "/openni_rgb_optical_frame";
    NODELET_INFO ("'rgb_frame_id' not set. using default: '%s'", rgb_frame_id_.c_str());
  }
  else
    NODELET_INFO ("rgb_frame_id = '%s' ", rgb_frame_id_.c_str());

  param_nh.param ("depth_frame_id", depth_frame_id_, string (""));
  if (depth_frame_id_.empty ())
  {
    depth_frame_id_ = "/openni_depth_optical_frame";
    NODELET_INFO ("'depth_frame_id' not set. using default: '%s'", depth_frame_id_.c_str());
  }
  else
    NODELET_INFO ("depth_frame_id = '%s' ", depth_frame_id_.c_str());

  param_nh.param ("use_indices", use_indices_, false);

	// setup image subscription call back
	assert(0);

  image_transport::ImageTransport imageTransport (comm_nh);
  image_transport::SubscriberStatusCallback subscriberChanged = boost::bind(&OpenNIRecordPlayerNodelet::subscriberChangedEvent, this);
  pub_rgb_image_   = imageTransport.advertise ("rgb/image_color", 5, subscriberChanged, subscriberChanged );
  pub_gray_image_  = imageTransport.advertise ("rgb/image_mono" , 5, subscriberChanged, subscriberChanged );
  pub_depth_image_ = imageTransport.advertise ("depth/image"    , 5, subscriberChanged, subscriberChanged );

  ros::SubscriberStatusCallback subscriberChanged2 = boost::bind(&OpenNIRecordPlayerNodelet::subscriberChangedEvent, this);
  pub_disparity_ = comm_nh.advertise<stereo_msgs::DisparityImage > ("depth/disparity", 5, subscriberChanged2, subscriberChanged2);
  pub_point_cloud_ = comm_nh.advertise<pcl::PointCloud<pcl::PointXYZ> > ("depth/points", 5, subscriberChanged2, subscriberChanged2);
  pub_point_cloud_rgb_ = comm_nh.advertise<pcl::PointCloud<pcl::PointXYZRGB> > ("rgb/points", 5, subscriberChanged2, subscriberChanged2);

  if (use_indices_)
    sub_mask_indices_ = comm_nh.subscribe("depth/indices", 5, &OpenNIRecordPlayerNodelet::maskIndicesCb, this);

  SyncPolicy sync_policy (4); // queue size
  depth_rgb_sync_.reset (new Synchronizer (sync_policy));
  depth_rgb_sync_->registerCallback (boost::bind (&OpenNIRecordPlayerNodelet::publishXYZRGBPointCloud, this, _1, _2));

  // initialize dynamic reconfigure
  reconfigure_server_.reset (new ReconfigureServer (reconfigure_mutex_, param_nh));
  reconfigure_mutex_.lock ();
  reconfigure_server_->updateConfig (config_);
  reconfigure_server_->setCallback (boost::bind (&OpenNIRecordPlayerNodelet::configCallback, this, _1, _2));
  reconfigure_mutex_.unlock ();
}

void OpenNIRecordPlayerNodelet::imageCallback (boost::shared_ptr<openni_wrapper::Image> image, void* cookie)
{
  ros::Time time = ros::Time::now () + ros::Duration(config_.image_time_offset);

  // mode is switching -> probably image sizes are not consistend... skip this frame
  //if (!image_mutex_.try_lock ())
  //  return;

  if (pub_rgb_info_.getNumSubscribers () > 0)
    pub_rgb_info_.publish (fillCameraInfo (time, true));

  if (pub_image_raw_.getNumSubscribers () > 0)
    publishRgbImageRaw (*image, time);

  if (pub_rgb_image_.getNumSubscribers () > 0 || pub_point_cloud_rgb_.getNumSubscribers () > 0 )
    publishRgbImage (*image, time);

  if (pub_gray_image_.getNumSubscribers () > 0)
    publishGrayImage (*image, time);

  //image_mutex_.unlock ();
}

void OpenNIRecordPlayerNodelet::subscriberChangedEvent ()
{
}

void OpenNIRecordPlayerNodelet::publishRgbImage (const Image& image, ros::Time time) const
{
  sensor_msgs::ImagePtr rgb_msg = boost::make_shared<sensor_msgs::Image > ();
  rgb_msg->header.stamp = time;
  rgb_msg->header.frame_id = rgb_frame_id_;
  rgb_msg->encoding = sensor_msgs::image_encodings::RGB8;
  rgb_msg->height = image_height_;
  rgb_msg->width = image_width_;
  rgb_msg->step = image_width_ * 3;
  rgb_msg->data.resize (rgb_msg->height * rgb_msg->step);
  image.fillRGB (rgb_msg->width, rgb_msg->height, &rgb_msg->data[0], rgb_msg->step);

  if (pub_rgb_image_.getNumSubscribers () > 0)
    pub_rgb_image_.publish (rgb_msg);
  
  if (pub_point_cloud_rgb_.getNumSubscribers () > 0)
    depth_rgb_sync_->add < 1 > (rgb_msg);
}

void OpenNIRecordPlayerNodelet::publishGrayImage (const Image& image, ros::Time time) const
{
  sensor_msgs::ImagePtr gray_msg = boost::make_shared<sensor_msgs::Image > ();
  gray_msg->header.stamp = time;
  gray_msg->header.frame_id = rgb_frame_id_;
  gray_msg->encoding = sensor_msgs::image_encodings::MONO8;
  gray_msg->height = image_height_;
  gray_msg->width = image_width_;
  gray_msg->step = image_width_;
  gray_msg->data.resize (gray_msg->height * gray_msg->step);
  image.fillGrayscale (gray_msg->width, gray_msg->height, &gray_msg->data[0], gray_msg->step);

  pub_gray_image_.publish (gray_msg);
}

void OpenNIRecordPlayerNodelet::publishDepthImage (const DepthImage& depth, ros::Time time) const
{
  sensor_msgs::ImagePtr depth_msg = boost::make_shared<sensor_msgs::Image > ();
  depth_msg->header.stamp         = time;
  depth_msg->header.frame_id      = device_->isDepthRegistered () ? rgb_frame_id_ : depth_frame_id_;
  depth_msg->encoding             = sensor_msgs::image_encodings::TYPE_32FC1;
  depth_msg->height               = depth_height_;
  depth_msg->width                = depth_width_;
  depth_msg->step                 = depth_msg->width * sizeof (float);
  depth_msg->data.resize (depth_msg->height * depth_msg->step);

  depth.fillDepthImage (depth_width_, depth_height_, reinterpret_cast<float*>(&depth_msg->data[0]), depth_msg->step);

  if (pub_depth_image_.getNumSubscribers () > 0)
    pub_depth_image_.publish (depth_msg);

  if (pub_point_cloud_rgb_.getNumSubscribers () > 0)
    depth_rgb_sync_->add < 0 > (depth_msg);
}

void OpenNIRecordPlayerNodelet::publishDisparity (const DepthImage& depth, ros::Time time) const
{
  stereo_msgs::DisparityImagePtr disp_msg = boost::make_shared<stereo_msgs::DisparityImage > ();
  disp_msg->header.stamp                  = time;
  disp_msg->header.frame_id               = device_->isDepthRegistered () ? rgb_frame_id_ : depth_frame_id_;
  disp_msg->image.header                  = disp_msg->header;
  disp_msg->image.encoding                = sensor_msgs::image_encodings::TYPE_32FC1;
  disp_msg->image.height                  = depth_height_;
  disp_msg->image.width                   = depth_width_;
  disp_msg->image.step                    = disp_msg->image.width * sizeof (float);
  disp_msg->image.data.resize (disp_msg->image.height * disp_msg->image.step);
  disp_msg->T = depth.getBaseline ();
  disp_msg->f = depth.getFocalLength () * depth_width_ / depth.getWidth ();

  /// @todo Compute these values from DepthGenerator::GetDeviceMaxDepth() and the like
  disp_msg->min_disparity = 0.0;
  disp_msg->max_disparity = disp_msg->T * disp_msg->f / 0.3;
  disp_msg->delta_d = 0.125;

  depth.fillDisparityImage (depth_width_, depth_height_, reinterpret_cast<float*>(&disp_msg->image.data[0]), disp_msg->image.step);
  pub_disparity_.publish (disp_msg);
}

void OpenNIRecordPlayerNodelet::publishXYZPointCloud (const DepthImage& depth, ros::Time time) const
{
  const xn::DepthMetaData& depth_md = depth.getDepthMetaData ();

  pcl::PointCloud<pcl::PointXYZ>::Ptr cloud_msg (new pcl::PointCloud<pcl::PointXYZ>() );
  cloud_msg->header.stamp = time;

  bool do_masking = (use_indices_ and (mask_indices_.size() != 0));

  float centerX, centerY;
  unsigned depthStep = 1, depthSkip = 0;
  if (not do_masking)
  {
    cloud_msg->height = depth_height_;
    cloud_msg->width = depth_width_;
    centerX = (cloud_msg->width >> 1) - 0.5f;
    centerY = (cloud_msg->height >> 1) - 0.5f;
    depthStep = depth_md.XRes() / cloud_msg->width;
    depthSkip = (depth_md.YRes() / cloud_msg->height - 1) * depth_md.XRes();
  }
  else
  {
    cloud_msg->width = mask_indices_.size();
    cloud_msg->height = 1;
    centerX = (depth_width_ >> 1) - 0.5f;
    centerY = (depth_height_ >> 1) - 0.5f;
  }
  cloud_msg->is_dense     = false;

  cloud_msg->points.resize (cloud_msg->height * cloud_msg->width);

  float constant = 0.001 / device_->getDepthFocalLength (depth_width_);

  if (device_->isDepthRegistered ())
    cloud_msg->header.frame_id = rgb_frame_id_;
  else
    cloud_msg->header.frame_id = depth_frame_id_;

  float bad_point = std::numeric_limits<float>::quiet_NaN ();

  int depth_idx = 0;
  pcl::PointCloud<pcl::PointXYZ>::iterator pt_iter = cloud_msg->begin ();
  if (not do_masking)
  {
    for (int v = 0; v < (int)cloud_msg->height; ++v, depth_idx += depthSkip)
    {
      for (int u = 0; u < (int)cloud_msg->width; ++u, depth_idx += depthStep, ++pt_iter)
      {
        pcl::PointXYZ& pt = *pt_iter;

        // Check for invalid measurements
        if (depth_md[depth_idx] == 0 ||
            depth_md[depth_idx] == depth.getNoSampleValue () ||
            depth_md[depth_idx] == depth.getShadowValue ())
        {
          // not valid
          pt.x = pt.y = pt.z = bad_point;
          continue;
        }

        // Fill in XYZ
        pt.x = (u - centerX) * depth_md[depth_idx] * constant;
        pt.y = (v - centerY) * depth_md[depth_idx] * constant;
        pt.z = depth_md[depth_idx] * 0.001;
      }
    }
  }
  else
  {
    int u, v;
    unsigned int nMask = mask_indices_.size();
    int max_index = (depth_height_ * depth_width_) - 1;
    for (unsigned int i = 0; i < nMask; i++, ++pt_iter)
    {
      pcl::PointXYZ& pt = *pt_iter;

      depth_idx = mask_indices_[i];
      if (depth_idx > max_index)
      {
        NODELET_ERROR("Mask index %d exceeds maximum index of %d", depth_idx, max_index);
        continue;
      }
      v = depth_idx / depth_width_;
      u = depth_idx % depth_width_;

      // Check for invalid measurements
      if (depth_md[depth_idx] == 0 ||
          depth_md[depth_idx] == depth.getNoSampleValue () ||
          depth_md[depth_idx] == depth.getShadowValue ())
      {
        // not valid
        pt.x = pt.y = pt.z = bad_point;
        continue;
      }

      // Fill in XYZ
      pt.x = (u - centerX) * depth_md[depth_idx] * constant;
      pt.y = (v - centerY) * depth_md[depth_idx] * constant;
      pt.z = depth_md[depth_idx] * 0.001;
    }
  }

  pub_point_cloud_.publish (cloud_msg);
}

void OpenNIRecordPlayerNodelet::publishXYZRGBPointCloud (const sensor_msgs::ImageConstPtr& depth_msg, const sensor_msgs::ImageConstPtr& rgb_msg) const
{
  pcl::PointCloud<pcl::PointXYZRGB>::Ptr cloud_msg (new pcl::PointCloud<pcl::PointXYZRGB>() );
  cloud_msg->header.stamp     = depth_msg->header.stamp;
  cloud_msg->header.frame_id  = rgb_frame_id_;
  cloud_msg->is_dense         = false;

  bool do_masking = (use_indices_ and (mask_indices_.size() != 0));

  float centerX, centerY;
  float constant;
  unsigned color_step, color_skip;

  if (not do_masking)
  {
    cloud_msg->height = depth_msg->height;
    cloud_msg->width = depth_msg->width;
    centerX = (cloud_msg->width >> 1) - 0.5f;
    centerY = (cloud_msg->height >> 1) - 0.5f;
    constant = 1.0f / device_->getImageFocalLength(cloud_msg->width);
    color_step = 3 * rgb_msg->width / cloud_msg->width;
    color_skip = 3 * (rgb_msg->height / cloud_msg->height - 1) * rgb_msg->width;

  }
  else
  {
    cloud_msg->width = mask_indices_.size();
    cloud_msg->height = 1;
    centerX = (depth_msg->width >> 1) - 0.5f;
    centerY = (depth_msg->height >> 1) - 0.5f;
    constant = 1.0f / device_->getImageFocalLength(depth_msg->width);
    color_step = 3 * rgb_msg->width / depth_msg->width;
    color_skip = 3 * (rgb_msg->height / depth_msg->height - 1) * rgb_msg->width;
 }

  static bool info_sent = false;
  if (not info_sent)
  {
    //NODELET_INFO("color_step = %d, color_skip = %d", color_step, color_skip);
    info_sent = true;
  }

  // do not publish if rgb image is smaller than color image -> seg fault
  if (rgb_msg->height < depth_msg->height || rgb_msg->width < depth_msg->width)
  {
    // we dont want to flood the terminal with warnings
    static unsigned warned = 0;
    if (warned % 100 == 0)
      NODELET_WARN("rgb image smaller than depth image... skipping point cloud for this frame rgb:%dx%d vs. depth:%3dx%d"
              , rgb_msg->width, rgb_msg->height, depth_msg->width, depth_msg->height );
    ++warned;
    return;
  }
  cloud_msg->points.resize (cloud_msg->height * cloud_msg->width);

  const float* depth_buffer = reinterpret_cast<const float*>(&depth_msg->data[0]);
  const uint8_t* rgb_buffer = &rgb_msg->data[0];

  // depth_msg already has the desired dimensions, but rgb_msg may be higher res.
  int color_idx = 0, depth_idx = 0;
  pcl::PointCloud<pcl::PointXYZRGB>::iterator pt_iter = cloud_msg->begin ();
  if (not do_masking)
  {
    for (int v = 0; v < (int)cloud_msg->height; ++v, color_idx += color_skip)
    {
      for (int u = 0; u < (int)cloud_msg->width; ++u, color_idx += color_step, ++depth_idx, ++pt_iter)
      {
        pcl::PointXYZRGB& pt = *pt_iter;
        float Z = depth_buffer[depth_idx];

        // Check for invalid measurements
        if (std::isnan (Z))
        {
          pt.x = pt.y = pt.z = Z;
        }
        else
        {
          // Fill in XYZ
          pt.x = (u - centerX) * Z * constant;
          pt.y = (v - centerY) * Z * constant;
          pt.z = Z;
        }

        // Fill in color
        RGBValue color;
        color.Red   = rgb_buffer[color_idx];
        color.Green = rgb_buffer[color_idx + 1];
        color.Blue  = rgb_buffer[color_idx + 2];
        color.Alpha = 0;
        pt.rgb = color.float_value;
      }
    }
  }
  else
  {
    int u, v;
    unsigned int nMask = mask_indices_.size();
    int max_index = (depth_msg->height * depth_msg->width) - 1;
    for (unsigned int i = 0; i < nMask; i++, ++pt_iter)
    {
      pcl::PointXYZRGB& pt = *pt_iter;
      depth_idx = mask_indices_[i];
      if (depth_idx > max_index)
      {
        NODELET_ERROR("Mask index %d exceeds maximum index of %d", depth_idx, max_index);
        continue;
      }
      float Z = depth_buffer[depth_idx];

      v = depth_idx / depth_msg->width;
      u = depth_idx % depth_msg->width;
      // Check for invalid measurements

      if (std::isnan (Z))
      {
        pt.x = pt.y = pt.z = Z;
      }
      else
      {
        // Fill in XYZ
        pt.x = (u - centerX) * Z * constant;
        pt.y = (v - centerY) * Z * constant;
        pt.z = Z;
      }

      // Fill in color
      RGBValue color;
      color_idx = (v*depth_msg->width + u)*color_step;
      color.Red   = rgb_buffer[color_idx];
      color.Green = rgb_buffer[color_idx + 1];
      color.Blue  = rgb_buffer[color_idx + 2];
      color.Alpha = 0;
      pt.rgb = color.float_value;
    }
  }

  pub_point_cloud_rgb_.publish (cloud_msg);
}

void OpenNIRecordPlayerNodelet::configCallback (Config &config, uint32_t level)
{
  config_ = config;
}

void OpenNIRecordPlayerNodelet::maskIndicesCb(const pcl::PointIndicesConstPtr& indices)
{
  mask_indices_.resize(indices->indices.size());
//  NODELET_INFO("New mask with %d indices", int(mask_indices_.size()));
  memcpy(mask_indices_.data(), indices->indices.data(), indices->indices.size() * sizeof(int32_t));
}
}


