/*
 * segmenter.cc
 * Mac Mason <mmason@willowgarage.com>
 *
 * Implementation. See segmenter.hh.
 */

#include <sstream>
#include <boost/foreach.hpp>
#define foreach BOOST_FOREACH
#include <boost/format.hpp>

#include "ros/ros.h"
#include "pcl_ros/transforms.h"
#include "pcl/filters/passthrough.h"
#include "pcl/filters/voxel_grid.h"
#include "pcl/sample_consensus/method_types.h"
#include "pcl/sample_consensus/model_types.h"
#include "pcl/filters/project_inliers.h"
#include "pcl/filters/extract_indices.h"
#include "pcl/filters/radius_outlier_removal.h"
#include "pcl/segmentation/sac_segmentation.h"
#include "pcl/segmentation/extract_clusters.h"
#include "pcl/segmentation/extract_polygonal_prism_data.h"
#include "pcl/kdtree/kdtree.h"
#include "pcl/surface/convex_hull.h"
#include "pcl/ros/conversions.h"

#include "semanticmodel/Plane.h"
#include "semanticmodel/centroid.hh"
#include "semanticmodel/segmenter.hh"
#include "semanticmodel/Blobs.h"
#include "semanticmodel/PlaneExchange.h"

// These are for debugging; I should get rid of them.
std::string CloudsToString(
    const std::vector< pcl::PointCloud<pcl::PointXYZRGB>::Ptr >& vec)
{
  std::stringstream ss;
  ss << vec.size() << ": [ ";
  for (size_t i = 0 ; i < vec.size() ; ++i)
  {
    ss << vec[i]->size() << " ";
  }
  ss << "]";
  return ss.str();
}

double degrees(double radians)
{
  return radians * 180.0 / M_PI;
}

// See above, re: debugging.
class ScopedTickTock
{
  public:
    std::string start;
    std::string end;
    ros::Time starttime;
    ScopedTickTock(const std::string& _start, const std::string& _end)
      : start(_start), end(_end), starttime(ros::Time::now())
    {
      ROS_INFO_STREAM("TickTock Start: " << start);
    }
    ~ScopedTickTock()
    {
      ROS_INFO_STREAM("TickTock Stop: " << 
                      end << 
                      " (" << 
                      (ros::Time::now() - starttime).toSec() <<
                      ") sec");

    }
};

namespace semanticmodel
{

Segmenter::Segmenter(const std::string& name_space,
                     const std::string& cloud_input_topic,
                     const std::string& camera_input_topic,
                     const std::string& planecloud_output_topic,
                     const std::string& plane_output_topic,
                     const std::string& object_output_topic,
                     const std::string& plane_service_topic)
  : handle(name_space),
    config(),
    dynparam_srv(),
    dynparam_func(),
    config_mutex(),
    listener(NULL),
    broadcaster(NULL),
    sync(NULL),
    cloud_sub(),
    transport(handle),
    cam_sub(),
    planecloud_pub(),
    plane_pub(),
    object_pub(),
    planetracker_srv(),
    short_circuit(false),
    processed(0),
    skipped(0)
{
  // Enforce our input requirements.
  ROS_FATAL_COND(
      cloud_input_topic == "" || camera_input_topic == "",
      "You must provide both cloud_input_topic and image_input_topic!");
  ROS_ASSERT(cloud_input_topic != "" && camera_input_topic != "");

  // Set up dynamic_reconfigure.
  dynparam_func = 
    boost::bind(&Segmenter::dynamic_reconfigure_callback, this, _1, _2);
  dynparam_srv.setCallback(dynparam_func);

  listener.reset(new tf::TransformListener());
  broadcaster.reset(new tf::TransformBroadcaster());

  // See if we're short-circuiting.
  if (!handle.getParam("/short_circuit", short_circuit))
    ROS_WARN("short_circuit parameter is missing; using false.");
  else
    ROS_WARN_STREAM("Setting short_circuit to " << short_circuit);

  // Get the synchronizer spun up.
  sync.reset(new message_filters::Synchronizer<Sync>(Sync(QUEUE_SIZE)));
  sync->registerCallback(boost::bind(
                           &Segmenter::synchronized_pipeline_callback, 
                           this, _1, _2, _3));

  // Subscribe to the topics we want.
  cloud_sub = handle.subscribe(cloud_input_topic, 
                               QUEUE_SIZE, 
                               &Segmenter::cloud_cb,
                               this);
  cam_sub = transport.subscribeCamera(camera_input_topic, 
                                      QUEUE_SIZE,
                                      &Segmenter::camera_cb,
                                      this);
  // Advertise our outputs.
  planecloud_pub = 
    handle.advertise<PointCloud>(planecloud_output_topic, QUEUE_SIZE);
  plane_pub = 
    handle.advertise<Plane>(plane_output_topic, QUEUE_SIZE);
  object_pub = handle.advertise<Blobs>(object_output_topic, QUEUE_SIZE);

  planetracker_srv = 
    handle.serviceClient<semanticmodel::PlaneExchange>(plane_service_topic);
}

void Segmenter::passthrough(PointCloud::Ptr& in, PointCloud::Ptr& out)
{
  pcl::PassThrough<Point> passfilter;
  passfilter.setFilterFieldName("z");
  passfilter.setFilterLimits(config.min_z, config.max_z);
  passfilter.setInputCloud(in);
  passfilter.filter(*out);
}

void Segmenter::voxelize(PointCloud::Ptr& in, PointCloud::Ptr& out)
{
  pcl::VoxelGrid<Point> voxelfilter;
  voxelfilter.setLeafSize(config.voxel_size, 
                          config.voxel_size, 
                          config.voxel_size);
  voxelfilter.setInputCloud(in);
  voxelfilter.filter(*out);
}

void Segmenter::fit_planes(
    PointCloud::Ptr& in, 
    std::vector<PointCloud::Ptr>& good_out,
    std::vector<pcl::ModelCoefficients::Ptr>& good_models,
    std::vector<pcl::ModelCoefficients::Ptr>& bad_models)
{
  // Finding planes.
  pcl::SACSegmentation<Point> segmenter;
  segmenter.setOptimizeCoefficients(true);
  segmenter.setModelType(pcl::SACMODEL_PLANE);
  segmenter.setMethodType(pcl::SAC_RANSAC);
  segmenter.setDistanceThreshold(config.plane_tolerance);
  segmenter.setMaxIterations(config.ransac_iterations);

  double orig_size = in->points.size();
  // TODO TODO or 0.5 could be a non-magic number, that could work too.
  int loops = 0;
  while (in->points.size() > orig_size * 0.5 && ros::ok())
  {
    // Our hypothetical cloud, model, and indices.
    PointCloud::Ptr cloud(new PointCloud());
    pcl::ModelCoefficients::Ptr model(new pcl::ModelCoefficients());
    pcl::PointIndices::Ptr indices(new pcl::PointIndices());

    // We're going to need this, no matter what.
    pcl::ExtractIndices<Point> extractor;

    // VERB!
    segmenter.setInputCloud(in);
    segmenter.segment(*indices, *model);

    if (indices->indices.size() == 0)
      break;  // We stop as soon as plane-fitting has trouble.

    // We want the dot-product between the plane normal (model.values) and the
    // +z axis (0 0 1); that's actually just the z-component of the normal,
    // because of the zeros.
    if (model->values[2] > config.horizontal_dot)
    {
      // Get those points out.
      extractor.setInputCloud(in);
      extractor.setNegative(false);
      extractor.setIndices(indices);
      extractor.filter(*cloud);
      Eigen::Vector4f middle = semanticmodel::centroid(cloud);
      foreach(Point& p, cloud->points)
        p.z = middle[2];

      model->values[0] = 0.0;
      model->values[1] = 0.0;
      model->values[2] = 1.0;
      model->values[3] = -middle[2];
      // Project the plane.
      pcl::ProjectInliers<Point> projector;
      projector.setModelType(pcl::SACMODEL_PLANE);
      projector.setInputCloud(cloud);
      projector.setModelCoefficients(model);
      projector.filter(*cloud);
      
      // Store the result.
      good_out.push_back(cloud);
      good_models.push_back(model);
    }
    else
    {
      bad_models.push_back(model);
    }

    // Horizontal or not, time to move on to a different plane.
    PointCloud::Ptr temp(new PointCloud());
    extractor.setInputCloud(in);
    extractor.setIndices(indices);
    extractor.setNegative(true);
    extractor.filter(*temp);
    std::swap(temp, in);
    loops++;
  }
}

void Segmenter::erode(std::vector<PointCloud::Ptr>& planes)
{
  // We'll fill this in, and then swap it with planes.
  std::vector<PointCloud::Ptr> output;

  // Neighbor searching.
  foreach(PointCloud::Ptr& plane, planes)
  {
    pcl::RadiusOutlierRemoval<Point> knn;
    knn.setRadiusSearch(config.k_radius);
    knn.setMinNeighborsInRadius(config.k_n);
    knn.setInputCloud(plane);
    knn.filter(*plane);
    output.push_back(plane);  // This might have size 0; see the caller.
  }
  std::swap(planes, output);
}

void Segmenter::plane_clusters(std::vector<PointCloud::Ptr>& planes,
                               std::vector<pcl::ModelCoefficients::Ptr>& coeffs)
{
  ROS_ASSERT(planes.size() == coeffs.size());
  pcl::ExtractIndices<Point> extractor;

  // We'll swap these with their originals when we're done.
  std::vector<PointCloud::Ptr> newplanes;
  std::vector<pcl::ModelCoefficients::Ptr> newcoeffs;

  // And go.
  for (size_t i = 0 ; i < planes.size() ; ++i)
  {
    pcl::KdTree<Point>::Ptr tree(new pcl::KdTreeFLANN<Point>());
    pcl::EuclideanClusterExtraction<Point> euclid;
    euclid.setClusterTolerance(config.plane_cluster_tolerance);
    euclid.setMinClusterSize(config.plane_min_cluster_size);
    euclid.setMaxClusterSize(config.plane_max_cluster_size);
    tree->setInputCloud(planes[i]);
    euclid.setSearchMethod(tree);
    euclid.setInputCloud(planes[i]);

    std::vector<pcl::PointIndices> clusters;
    euclid.extract(clusters);
    extractor.setInputCloud(planes[i]);

    foreach(pcl::PointIndices& cluster, clusters)
    {
      PointCloud::Ptr cloud(new PointCloud());
      // Stupid PCL memory-management choices.
      pcl::PointIndices::Ptr pcluster(new pcl::PointIndices(cluster));
      extractor.setIndices(pcluster);
      extractor.setNegative(false);
      extractor.filter(*cloud);
      newplanes.push_back(cloud);

      // Copy the model from the original cloud.
      newcoeffs.push_back(coeffs[i]);
    }
  }
  std::swap(newplanes, planes);
  std::swap(newcoeffs, coeffs);
}

void Segmenter::convex_hulls(const std::vector<PointCloud::Ptr>& planes,
                             std::vector<PointCloud::Ptr>& hulls)
{
  pcl::ConvexHull<Point> hull;
  foreach(const PointCloud::Ptr& plane, planes)
  {
    PointCloud::Ptr out(new PointCloud());
    hull.setInputCloud(plane);
    hull.reconstruct(*out);
    hulls.push_back(out);
  }
}

void Segmenter::publish_planes(
    const std::vector<PointCloud::Ptr>& planes,
    const std::vector<PointCloud::Ptr>& hulls,
    const std::vector<pcl::ModelCoefficients::Ptr>& coeffs,
    std::vector<Plane>& outputs)
{
  ROS_ASSERT(planes.size() == hulls.size());
  ROS_ASSERT(hulls.size() == coeffs.size());
  // Pushing the plane messages is easy; the tricky part is that we also want
  // to push a tf frame for best_frame, for the benefit of the head pointer.
  // So, we need to know which one is the biggest before we start.
  //
  // At the moment, we're going to publish really boring names for the planes;
  // just the index of each plane.
  if (planes.size() == 0)
    return;

  size_t largest = planes[0]->points.size();
  size_t largest_idx = 0;
  for (size_t i = 0 ; i < planes.size() ; ++i)
  {
    if (planes[i]->points.size() > largest)
    {
      largest = planes[i]->points.size();
      largest_idx = i;
    }
  }

  // Now we have the index of the best plane; let's publish everything.
  PlaneExchange exchange;
  exchange.request.in_planes.planes.resize(planes.size());
  for (size_t i = 0 ; i < planes.size() ; ++i)
  {
    Plane planemsg;
    planemsg.header = planes[i]->header;

    planemsg.a = coeffs[i]->values[0];
    planemsg.b = coeffs[i]->values[1];
    planemsg.c = coeffs[i]->values[2];
    planemsg.d = coeffs[i]->values[3];

    Eigen::Vector4f center = centroid(planes[i]);
    planemsg.center.x = center[0];
    planemsg.center.y = center[1];
    planemsg.center.z = center[2];

    pcl::toROSMsg(*hulls[i], planemsg.hull);
    pcl::toROSMsg(*planes[i], planemsg.cloud);

    exchange.request.in_planes.planes[i] = planemsg;
    
    // planemsg is built and stored. Now, build the tf transform.
    tf::Transform transform;
    transform.setOrigin(tf::Vector3(center[0], center[1], center[2]));
    transform.setRotation(tf::Quaternion(0, 0, 1.0, 0.0));
    std::string name = (boost::format("plane_%02d") % i).str();
    tf::StampedTransform trans(transform, 
                               planes[i]->header.stamp,
                               config.canonical_frame,
                               name);
    broadcaster->sendTransform(trans);
    
    // An extra one for the best current plane.
    if (i == largest_idx)
    {
      tf::StampedTransform trans2(transform, 
                                  planes[i]->header.stamp,
                                  config.canonical_frame,
                                  "best_plane");
      broadcaster->sendTransform(trans2);
    }
  }
  // Now, transmit our planes, and get the new hulls back.
  planetracker_srv.call(exchange);
  outputs = exchange.response.out_planes.planes;
}

void Segmenter::get_aboves(
    const std::vector<PointCloud::Ptr>& hulls,
    const std::vector<pcl::ModelCoefficients::Ptr>& coeffs,
    const PointCloud::Ptr& densecloud,
    PointCloud::Ptr& out)
{
  out->header = densecloud->header;
  foreach(Point& p, *densecloud)
  {
    // We want to use every point that's "above enough" a plane, but isn't
    // _below_ any planes at all. That's not perfect, but it's a start.
    bool add = false;
    for (size_t i = 0 ; i < hulls.size() ; ++i)
    {
      // Point-to-plane distance.
      double ppdist = coeffs[i]->values[0] * p.x + 
                      coeffs[i]->values[1] * p.y + 
                      coeffs[i]->values[2] * p.z + 
                      coeffs[i]->values[3];
      if (hulls[i]->points.size()>3 && pcl::isXYPointIn2DXYPolygon(p, *hulls[i]))
      {
        if (ppdist > config.plane_distance)
        {
          add = true;
        }
        else
        {
          add = false;
          break;
        }
      }
    }
    if (add)
    {
      out->points.push_back(p);
    }
  }
}

void Segmenter::lose_bads(const PointCloud::ConstPtr& aboves,
                          PointCloud::Ptr good_aboves,
                          const vector<pcl::ModelCoefficients::Ptr>& bad_coeffs)
{
  good_aboves->header = aboves->header;
  foreach(const Point& p, *aboves)
  {
    bool add = true;
    foreach(pcl::ModelCoefficients::Ptr model, bad_coeffs)
    {
      double ppdist = model->values[0] * p.x + 
                      model->values[1] * p.y + 
                      model->values[2] * p.z + 
                      model->values[3];
      if (fabs(ppdist) <= config.plane_tolerance)  // Fits the plane.
      {
        add = false;
        break;
      }
    }
    if (add)
      good_aboves->points.push_back(p);
  }
}

void Segmenter::synchronized_pipeline_callback(
    const sensor_msgs::ImageConstPtr& img,
    const sensor_msgs::CameraInfoConstPtr& info,
    const PointCloud::ConstPtr& cloud)
{
  // No changes during a frame, please.
  boost::mutex::scoped_lock lock(config_mutex);

  if (processed + skipped > 0)
  {
    ROS_DEBUG_NAMED("dropped_frames",
                    "Processed %d frames. Skipped %d frames. (%f processed)",
                    processed, 
                    skipped, 
                    ((float)processed) / (processed + skipped));
  }
  
  ROS_DEBUG_STREAM_NAMED("dropped_frames", "We are behind by "
                         << ros::Time::now() - img->header.stamp);
  // Avoid falling too far back
  if (img->header.stamp + ros::Duration(config.filter_lag_cutoff) < 
      ros::Time::now())
  {
    ROS_WARN_STREAM_NAMED("filter", 
                          "Too old! (" << 
                          ros::Time::now() - img->header.stamp << ")");
    skipped++;
    return;
  }
  else
  {
    ROS_DEBUG_STREAM_NAMED("filter", 
                           "Processing filter input at " << img->header.stamp);
    processed++;
  }


  // Step 0:
  // We get the cloud by ConstPtr, so we may as well start with a copy. PCL's
  // inside-out concept of where the shared_ptr should go means that keeping
  // everything in a ::Ptr is probably easiest.
  PointCloud::Ptr transformed(new PointCloud()); 
  pcl_ros::transformPointCloud(config.canonical_frame,
                               *cloud, *transformed, *listener);

  // Step 1:
  // PCL does a woeful job of explaining when it's ok to use the same cloud
  // as both input and output for a filter. So, we end up with some extraneous
  // point clouds.
  PointCloud::Ptr passedthrough(new PointCloud());
  passthrough(transformed, passedthrough);
  if (passedthrough->points.size() == 0 || !ros::ok())
  {
    ROS_DEBUG_STREAM_NAMED("filter", "No passed-through points; skipping");
    return;
  }
  transformed.reset();  // Never going to use this one again.

  // Step 2:
  // The size into which you are voxelizing is seriously important re: speed.
  // One cm seems pretty solid. Four cm is much, much faster, and works too.
  PointCloud::Ptr voxelized(new PointCloud());
  voxelize(passedthrough, voxelized);
  if (voxelized->points.size() == 0 || !ros::ok())
  {
    ROS_WARN_STREAM_NAMED("filter", 
                          "No voxelized points; skipping (this is unexpected)");
    return;
  }

  // Step 3:
  // This is the slow part...
  std::vector<PointCloud::Ptr> good_planes;
  std::vector<pcl::ModelCoefficients::Ptr> good_coeffs;
  std::vector<PointCloud::Ptr> bad_planes;
  std::vector<pcl::ModelCoefficients::Ptr> bad_coeffs;
  fit_planes(voxelized, good_planes, good_coeffs, bad_coeffs);

  if (good_planes.size() == 0 || !ros::ok())
  {
    ROS_DEBUG_STREAM_NAMED("filter", "No planes found; skipping");
    return;
  }

  voxelized.reset();  // May as well release here as anywhere.

  /*
   * Steps 4 and 4.5 have been commented out because (with the 6cm
   * voxelization) the problem they solve seems have evaporated. Handy!
   */
  // Step 4:
  // The other slow part. 
  // Commented out; with the new 5cm voxelization, no longer seems to be
  // needed. (This may also be the result of pointing the head at the floor;
  // at any rate, we're faster without it)
  //erode(planes);

  // Step 4.5:
  // erode() might have driven some planes right down to zero points, or a
  // value so small as to not count; in this case, we need to get rid of their
  // coefficients, too, so that the coefficients and planes are still
  // index-aligned.
//  std::vector<PointCloud::Ptr> newplanes;
//  std::vector<pcl::ModelCoefficients::Ptr> newcoeffs;
//  for (size_t i = 0 ; i < planes.size() ; ++i)
//  {
//    if (planes[i]->points.size() < size_t(config.plane_min_cluster_size))
//      continue;
//    newplanes.push_back(planes[i]);
//    newcoeffs.push_back(coeffs[i]);
//  }
//  std::swap(newcoeffs, coeffs);
//  std::swap(newplanes, planes);
//  newcoeffs.clear();  // No need for two copies anymore.
//  newplanes.clear();

//  if (good_planes.size() == 0 || !ros::ok())
//  {
//    ROS_DEBUG_STREAM_NAMED("filter", "No planes after erosion; skipping");
//    return;
//  }

  // Step 5: We are likely to end up with even more planes after this step,
  // but we might actually end up with 0.
  plane_clusters(good_planes, good_coeffs);
  if (good_planes.size() == 0 || !ros::ok())
  {
    ROS_DEBUG_STREAM_NAMED("filter", "No planes after clustering; skipping");
    return;
  }

  // Step 6. Convex hulls. We need to keep the plane points, so we can't
  // clobber the planes variable.
  std::vector<PointCloud::Ptr> hulls;
  convex_hulls(good_planes, hulls); 
  if (!ros::ok())
    return;

  // Step 7. Send our planes to the tracker, and get back the new hulls.
  std::vector<Plane> returnedplanes;
  ROS_DEBUG_STREAM_NAMED("filter", "Calling publish_planes.");
  publish_planes(good_planes, hulls, good_coeffs, returnedplanes);
  if (short_circuit || !ros::ok())
    return;

  // Parse the newplanes out.
  hulls.clear();
  good_coeffs.clear();
  foreach(Plane& p, returnedplanes)
  {
    hulls.push_back(PointCloud::Ptr(new PointCloud()));
    pcl::fromROSMsg(p.hull, *hulls.back());
    good_coeffs.push_back(
        pcl::ModelCoefficients::Ptr(new pcl::ModelCoefficients()));
    good_coeffs.back()->values.resize(4);
    good_coeffs.back()->values[0] = p.a;
    good_coeffs.back()->values[1] = p.b;
    good_coeffs.back()->values[2] = p.c;
    good_coeffs.back()->values[3] = p.d;
  }

  // Step 8: we go back to working in the dense plane, here.
  PointCloud::Ptr aboves(new PointCloud());
  get_aboves(hulls, good_coeffs, passedthrough, aboves);
  if (aboves->points.size() == 0 || !ros::ok())
  {
    ROS_DEBUG_STREAM_NAMED("filter", "No above points; skipping");
    return;
  }

  // Step 8.5. Given that we have points that are "above enough" the
  // horizontal planes in the scene, we now want to throw out those that are
  // too good of a fit to the "bad" (non-horizontal) planes. The idea here is
  // to get rid of points on walls.
  PointCloud::Ptr good_aboves(new PointCloud());
  lose_bads(aboves, good_aboves, bad_coeffs);

  aboves.reset();

  if (good_aboves->points.size() == 0 || !ros::ok())
  {
    ROS_DEBUG_STREAM_NAMED("filter", "No good above points; skipping.");
    return;
  }

  /*
   * NEW VERSION.
   * Doing the extra object processing is expensive. Rather than do all of it,
   * we punt: send the simple aboves over the wire, and have the object
   * tracker keep it all together. We'd do lose_bads too, but it requires the
   * bad_coeffs variable, which I don't want to (also) ship over the wire.
   */
  Blobs msg;
  msg.blobs.resize(1);  // No clustering on this end.
  msg.camera_info = *info;
  msg.rgb_image = *img;
  pcl::toROSMsg(*good_aboves, msg.blobs.at(0));
  object_pub.publish(msg);
  return;

  //// Step 9: Objects!
  //std::vector<PointCloud::Ptr> clusters;
  //cluster_objects(good_aboves, clusters);

  //deplane_objects(clusters);
  //
  //// Step 0xA: Publish a single Blobs message for the objects
  //Blobs msg;
  //msg.blobs.resize(clusters.size());
  //msg.camera_info = *info;
  //msg.rgb_image = *img;
  //for (size_t i = 0 ; i < clusters.size() ; ++i)
  //  pcl::toROSMsg(*clusters.at(i), msg.blobs.at(i));

  //object_pub.publish(msg);
}

void Segmenter::dynamic_reconfigure_callback(const FilterConfig& config, 
                                             uint32_t level)
{
  // We ignore level, because everybody does.
  boost::mutex::scoped_lock lock(config_mutex);
  this->config = config;  // Well, that was easy.
}

// These next two functions just add stuff to sync's queue; it handles the
// rest.
void Segmenter::cloud_cb(const PointCloud::ConstPtr& cloud)
{
  sync->add<2>(cloud); 
}

void Segmenter::camera_cb(const sensor_msgs::ImageConstPtr& img, 
                          const sensor_msgs::CameraInfoConstPtr& info)
{
  sync->add<0>(img);
  sync->add<1>(info);
}

}  // namespace semanticmodel
